"""
Provides functions that are used to identify and reduce the set of attractor candidate states.

Note that most of these functions assume they are operating on the network or petri net
percolated to the subspace of the relevant node.
"""

from __future__ import annotations

from typing import TYPE_CHECKING, Literal, cast

if TYPE_CHECKING:
    from biobalm.succession_diagram import SuccessionDiagram
    from biobalm.types import BooleanSpace
    from networkx import DiGraph  # type: ignore

import random
import biobalm
from biodivine_aeon import Bdd, AsynchronousGraph, BddVariable
from biobalm.trappist_core import compute_fixed_point_reduced_STG
from biobalm.symbolic_utils import state_list_to_bdd, valuation_to_state, state_to_bdd
from biobalm.space_utils import intersect, is_subspace

try:
    pint_available = True
    import biobalm._pint_reachability
except ModuleNotFoundError:
    pint_available = False


def compute_attractor_candidates(
    sd: SuccessionDiagram,
    node_id: int,
    greedy_asp_minification: bool,
    simulation_minification: bool,
    pint_minification: bool,
) -> list[BooleanSpace]:
    """
    Compute an optimized list of candidate states that is guaranteed
    to cover every attractor in the specified SD node.

    It should hold that for every attractor in the node's sub-space outside
    of the child sub-spaces (if known), the list contains at least one state
    from this attractor. However, it is not guaranteed that each candidate
    covers some attractor, so a non-empty list of candidates can
    still correspond to an empty list of attractors.

    The method starts by computing an NFVS of the network percolated to the
    node's space, and then assigns each NFVS node a value towards which
    it tends to. This transformation satisfies that the fixed-points in the
    resulting network cover the attractors in the original sub-space.

    Subsequently, the method can further optimize this set of candidates.

    Greedy ASP minification
    -----------------------
    Depending on which values are chosen for the NFVS nodes, a different set of
    candidates is obtained. This optimization method flips the initial NFVS node
    values and retains modifications that lead to a smaller candidate set, until
    a local minimum is achieved.

    Simulation minification
    -----------------------
    It is often possible to reduce the candidate set using a pseudo-random walk
    which eventually leads to one of the child spaces, or to one of the existing
    candidate states.

    Pint minification
    -----------------
    The tool `pint` can perform (incomplete) reachability check for any network
    state, hence we can also use it to reduce the list of candidate states.
    This option is only available if pint is actually installed.

    Parameters
    ----------
    sd : SuccessionDiagram
        The succession diagram in question.
    node_id : int
        An ID of an SD node for which we are computing the candidates.
    greedy_asp_minification: bool
        Whether to enable the iterative ASP-based minification.
    simulation_minification: bool
        Whether to enable simulation minification.
    pint_minification: bool
        Whether to enable pint minification.

    Returns
    -------
    list[BooleanSpace]
        The list of attractor candidate states.
    """

    if sd.config["debug"]:
        print(f"[{node_id}] Start computing attractor candidates.")

    node_data = sd.node_data(node_id)

    node_space = node_data["space"]

    if len(node_space) == sd.network.variable_count():
        if sd.config["debug"]:
            print(f"[{node_id}] > Attractor candidates done: node is a fixed-point.")
        return [node_space]

    node_nfvs = sd.node_percolated_nfvs(node_id, compute=True)

    if sd.config["debug"]:
        root_nfvs = sd.node_percolated_nfvs(sd.root(), compute=True)
        print(
            f"[{node_id}] > Percolated node NFVS contains {len(node_nfvs)} nodes (vs {len(root_nfvs)} in the root)."
        )

    # Compute the list of child stable-motifs if the node is expanded. Otherwise
    # "pretend" that there are no children.
    #
    # (We can safely ignore anything that is in these stable motifs, not just the
    # percolated child spaces, because it is either in the child space, or it percolates
    # to the child space, so it does not contain attractors)
    child_motifs_reduced = []
    if node_data["expanded"]:
        children = sd.node_successors(node_id, compute=False)
        child_motifs_reduced = [
            sd.edge_stable_motif(node_id, s, reduced=True) for s in children
        ]

    if node_data["skipped"]:
        # For skip nodes, it does not hold that the successors are the maximal subspaces.
        # This means that a skip node can intersect with some other SD node and that intersection
        # is not a subset of one of its children. In such case, we can use this intersection
        # to further simplify the attractor detection process.
        total_skip_nodes_applied = 0
        for n in sd.node_ids():
            n_data = sd.node_data(n)
            if is_subspace(node_space, n_data["space"]):
                # This means that (in a fully expanded SD), `node` would be a (transitive)
                # successor of `n`, which means that a result for `n` is "attractors in n
                # that are not in node". Hence, we can't use it to reason about `node`.
                # This is not a problem if the intersection of the two nodes is non-triviall,
                # because that means they have common successors (and those should be
                # solved separately), but the nodes themselves do not depend on each other.
                continue
            if n_data["attractor_candidates"] == [] or n_data["attractor_seeds"] == []:
                # This will create a lot of duplicates, but it seems to be better than
                # not doing it at all.
                common_subspace = intersect(node_space, n_data["space"])
                if common_subspace is not None:
                    reduced_subspace: BooleanSpace = {
                        k: v for k, v in common_subspace.items() if k not in node_space
                    }
                    child_motifs_reduced.append(reduced_subspace)
                    total_skip_nodes_applied += 1
        if sd.config["debug"]:
            print(
                f"[{node_id}] Extended child motifs with {total_skip_nodes_applied} skip-node intersections."
            )

    # Indicates that this space is either minimal, or has no computed successors.
    # In either case, the space must contain at least one attractor.
    node_is_pseudo_minimal = len(child_motifs_reduced) == 0

    if sd.config["debug"] and node_is_pseudo_minimal:
        print(
            f"[{node_id}] > The node has no children (i.e. it is minimal or unexpanded)."
        )

    if len(node_nfvs) == 0:
        # If the NFVS is empty, it means this node space has no complex attractors.
        # In a fully expanded SD, all fixed-points should be handled by the condition above,
        # so only non-minimal spaces remain and these can be resolved as empty.
        #
        # However, if SD is not fully expanded, it can contain further "pseudo-minimal"
        # spaces that would contain minimal traps if expanded. For those, there should
        # only be fixed-point attractors, but we need to identify
        # them, so we need to continue.

        # This tests the assumption that all fully expanded minimal
        # traps are handled before, and the only trap spaces with
        # an empty NFVS are unexpanded pseudo-minimal,
        # or expanded non-minimal.
        assert not sd.node_is_minimal(node_id)

        if not node_is_pseudo_minimal:
            if sd.config["debug"]:
                print(
                    f"[{node_id}] > Attractor candidates done: empty NFVS in a non-minimal space."
                )
            return []

    # First, we create a retained set and corresponding candidate set that is valid in the
    # percolated Boolean network. This process is *usually* quite fast, but
    # sometimes the candidate set explodes and is very large (even too large to enumerate).
    # In that case, we can try to simplify it greedily using the ASP solver (assuming this
    # is enabled).
    #
    # Overall, our strategy is the following:
    # 1. Compute a heuristic retained set based on the original NFVS method.
    # 2. If the ASP optimization is disabled, we generate the candidates from this retained
    #    set in full.
    # 3. If ASP optimization is enabled, we first test if there are at least 100 candidates
    #    for the initial retained set.
    #    3a. If there are fewer than 100 candidates, we just run one optimization pass on
    #        the initial retained set and report the result.
    #    3b. If there are more than 100 candidates, we probably need to optimize the retained
    #        set even further, in which case we start to recursively generate a new retained
    #        set that is optimized in each step to hopefully get a better result.

    pn_reduced = sd.node_percolated_petri_net(node_id, compute=True)
    bn_reduced = sd.node_percolated_network(node_id, compute=True)
    graph_reduced = AsynchronousGraph(bn_reduced)

    retained_set = make_heuristic_retained_set(
        graph_reduced, node_nfvs, child_motifs_reduced
    )

    if len(retained_set) == sd.network.variable_count() and node_is_pseudo_minimal:
        # If the retained set describes a fixed point, then only one attractor
        # is present in this space and it must contain the state described by the retained set.
        if sd.config["debug"]:
            print(
                f"[{node_id}] > Singular attractor found through fixed-point retained set. Done."
            )
        return [retained_set | node_space]

    if not greedy_asp_minification:
        candidate_states = compute_fixed_point_reduced_STG(
            pn_reduced,
            retained_set,
            avoid_subspaces=child_motifs_reduced,
            solution_limit=sd.config["attractor_candidates_limit"],
        )
        if len(candidate_states) == sd.config["attractor_candidates_limit"]:
            raise RuntimeError(
                f"Exceeded the maximum amount of attractor candidates ({sd.config['attractor_candidates_limit']}; see `SuccessionDiagramConfiguation.attractor_candidates_limit`)."
            )
        if sd.config["debug"]:
            print(
                f"[{node_id}] Computed {len(candidate_states)} candidate states without retained set optimization."
            )
    else:
        candidate_states = compute_fixed_point_reduced_STG(
            pn_reduced,
            retained_set,
            avoid_subspaces=child_motifs_reduced,
            solution_limit=sd.config["retained_set_optimization_threshold"],
        )

        if len(candidate_states) < sd.config["retained_set_optimization_threshold"]:
            # The candidate set is small. No need to overthink it. The original
            # retained set is probably quite good. However, we can still try to
            # optimize it further if it makes sense.
            if len(candidate_states) > 1 or (
                not node_is_pseudo_minimal and len(candidate_states) > 0
            ):
                if sd.config["debug"]:
                    print(
                        f"[{node_id}] Initial retained set generated {len(candidate_states)} candidates. Optimizing..."
                    )
                optimized = asp_greedy_retained_set_optimization(
                    sd,
                    node_id,
                    petri_net=pn_reduced,
                    retained_set=retained_set,
                    candidate_states=candidate_states,
                    avoid_dnf=child_motifs_reduced,
                )
                retained_set = optimized[0]
                candidate_states = optimized[1]
        else:
            # There seem to be many candidates, in which case it might be better
            # to optimize the retained set dynamically.
            if sd.config["debug"]:
                print(
                    f"[{node_id}] Initial retained set generated >{sd.config['retained_set_optimization_threshold']} candidates. Regenerate retained set."
                )
            retained_set = {}
            candidate_states = []
            for var in node_nfvs:
                retained_set[var] = 0
                candidate_states_zero = compute_fixed_point_reduced_STG(
                    pn_reduced,
                    retained_set,
                    avoid_subspaces=child_motifs_reduced,
                    solution_limit=sd.config["attractor_candidates_limit"],
                )

                if len(candidate_states_zero) <= len(candidate_states):
                    if sd.config["debug"]:
                        print(
                            f"[{node_id}] Chosen {var}=0 without increasing candidate count ({len(candidate_states_zero)}). {len(retained_set)}/{len(node_nfvs)} variables chosen."
                        )
                    candidate_states = candidate_states_zero
                    continue

                retained_set[var] = 1
                candidate_states_one = compute_fixed_point_reduced_STG(
                    pn_reduced,
                    retained_set,
                    avoid_subspaces=child_motifs_reduced,
                    solution_limit=len(candidate_states_zero),
                )

                if (
                    len(candidate_states_zero)
                    == sd.config["attractor_candidates_limit"]
                    and len(candidate_states_one)
                    == sd.config["attractor_candidates_limit"]
                ):
                    raise RuntimeError(
                        f"Exceeded the maximum amount of attractor candidates ({sd.config['attractor_candidates_limit']}; see `SuccessionDiagramConfiguation.attractor_candidates_limit`)."
                    )

                if len(candidate_states_one) <= len(candidate_states):
                    if sd.config["debug"]:
                        print(
                            f"[{node_id}] Chosen {var}=1 without increasing candidate count ({len(candidate_states_one)}). {len(retained_set)}/{len(node_nfvs)} variables chosen."
                        )
                    candidate_states = candidate_states_one
                    continue

                if len(candidate_states_zero) < len(candidate_states_one):
                    if sd.config["debug"]:
                        print(
                            f"[{node_id}] Chosen {var}=0 with better candidate count ({len(candidate_states_zero)}). {len(retained_set)}/{len(node_nfvs)} variables chosen."
                        )
                    candidate_states = candidate_states_zero
                    retained_set[var] = 0
                else:
                    if sd.config["debug"]:
                        print(
                            f"[{node_id}] Chosen {var}=1 with better candidate count ({len(candidate_states_one)}). {len(retained_set)}/{len(node_nfvs)} variables chosen."
                        )
                    candidate_states = candidate_states_one
                    retained_set[var] = 1

                # At this point, we know the candidate count increased and so we should
                # try to bring it back down.
                if (
                    len(candidate_states)
                    > sd.config["retained_set_optimization_threshold"]
                ):
                    if sd.config["debug"]:
                        print(f"[{node_id}] Optimizing partial retained set...")
                    optimized = asp_greedy_retained_set_optimization(
                        sd,
                        node_id,
                        petri_net=pn_reduced,
                        retained_set=retained_set,
                        candidate_states=candidate_states,
                        avoid_dnf=child_motifs_reduced,
                    )
                    retained_set = optimized[0]
                    candidate_states = optimized[1]

    # Terminate if done.
    if len(candidate_states) == 0:
        if sd.config["debug"]:
            print(f"[{node_id}] > Initial candidate set empty. Done.")
        return []
    if node_is_pseudo_minimal and len(candidate_states) == 1:
        if sd.config["debug"]:
            print(
                f"[{node_id}] > Single candidate found in (pseudo) minimal trap space. Done."
            )
        return [candidate_states[0] | node_space]

    if sd.config["debug"]:
        print(
            f"[{node_id}] > Attractor candidates from retained set: {len(candidate_states)}."
        )

    if simulation_minification:
        if sd.config["debug"]:
            print(f"[{node_id}] Start simulation minification...")
        avoid_children_symbolic = state_list_to_bdd(
            graph_reduced.symbolic_context(), child_motifs_reduced
        )

        # Here, we gradually increase the iteration count while
        # the candidate set is being actively reduced. If the simulation
        # cannot reduce any further states and exceeds the proposed budget,
        # we are done.
        iterations = 2**10
        max_budget = (
            sd.config["minimum_simulation_budget"] * bn_reduced.variable_count()
        )
        while len(candidate_states) > 0:
            if sd.config["debug"]:
                print(
                    f"[{node_id}] > Start simulation with {len(candidate_states)} states and simulation limit {iterations}."
                )
            reduced = run_simulation_minification(
                sd,
                node_id,
                graph_reduced,
                candidate_states,
                avoid_children_symbolic,
                max_iterations=iterations,
                simulation_seed=123,
            )

            if (
                len(reduced) == len(candidate_states)
                and (iterations * len(candidate_states)) > max_budget
            ):
                candidate_states = reduced
                break

            iterations = 2 * iterations
            candidate_states = reduced

            if len(candidate_states) == 1 and avoid_children_symbolic.is_false():
                break

        if sd.config["debug"]:
            print(f"[{node_id}] > Candidates after simulation: {len(candidate_states)}")

    # Terminate if done.
    if len(candidate_states) == 0:
        if sd.config["debug"]:
            print(f"[{node_id}] > Candidate set empty. Done.")
        return []
    if node_is_pseudo_minimal and len(candidate_states) == 1:
        if sd.config["debug"]:
            print(
                f"[{node_id}] > Single candidate found in (pseudo) minimal trap space. Done."
            )
        return [candidate_states[0] | node_space]

    if pint_minification and not pint_available:
        print("WARNING: Using `pint`, but `pint` is not installed. Skipping.")
    elif pint_minification and pint_available:
        if sd.config["debug"]:
            print(f"[{node_id}] Start `pint` minification...")

        children_bdd = state_list_to_bdd(
            graph_reduced.symbolic_context(), child_motifs_reduced
        )
        candidates_bdd = state_list_to_bdd(
            graph_reduced.symbolic_context(), candidate_states
        )
        avoid_bdd = children_bdd.l_or(candidates_bdd)

        filtered_states: list[BooleanSpace] = []
        for i, state in enumerate(candidate_states):
            state_bdd = state_to_bdd(graph_reduced.symbolic_context(), state)

            avoid_bdd = avoid_bdd.l_and_not(state_bdd)

            keep = True
            try:
                if biobalm._pint_reachability.pint_reachability(
                    pn_reduced, state, avoid_bdd, sd.config
                ):
                    keep = False
            except RuntimeError as e:
                assert str(e) == "Cannot verify."

            if keep:
                avoid_bdd = avoid_bdd.l_or(state_bdd)
                filtered_states.append(state)

            if sd.config["debug"]:
                print(
                    f"[{node_id}] > `pint` {i + 1}/{len(candidate_states)}: eliminated: {not keep}, retained: {len(filtered_states)}."
                )

        candidate_states = filtered_states

        if sd.config["debug"]:
            print(f"[{node_id}] > Candidates after `pint`: {len(candidate_states)}")

    # Finally, we need to augment each candidate state with the
    # fixed values from the node space, because until now, we only considered
    # the reduced/percolated network.
    candidate_states = [x | node_space for x in candidate_states]

    return candidate_states


def run_simulation_minification(
    sd: SuccessionDiagram,
    node_id: int,
    graph: AsynchronousGraph,
    candidate_states: list[BooleanSpace],
    avoid_bdd: Bdd,
    max_iterations: int,
    simulation_seed: int,
) -> list[BooleanSpace]:
    """
    A fast but incomplete method for eliminating spurious attractor candidates
    based on stochastic simulation.

    Parameters
    ----------
    node_id : int
        The ID of the associated SD node. This is only for logging progress.
    graph : biodivine_aeon.AsynchronousGraph
        The symbolic encoding of the *percolated* network dynamics.
    candidate_states: list[BooleanSpace]
        The list of currently considered candidates that is to be reduced.
    avoid_bdd: biodivine_aeon.Bdd
        A symbolic representation of the states/spaces that are to be ignored.
        If any of these states is reachable by a candidate, that candidate is
        safe to ignore as well.
    max_iterations: int
        The number of steps performed by the simulator.
    simulation_seed: int
        The seed value for the random walk simulator.

    Returns
    -------
    list[BooleanSpace]
        The optimized list of candidate states.
    """

    # A random generator initialized with a fixed seed. Ensures simulation
    # is randomized but deterministic.
    generator = random.Random(simulation_seed)

    # Retrieve the symbolic variables that encode each network variable and
    # associate them with the corresponding network's update functions.
    symbolic_ctx = graph.symbolic_context()
    network_vars = graph.network_variables()
    symbolic_vars: list[BddVariable] = []
    for var in network_vars:
        s_var = symbolic_ctx.find_network_bdd_variable(var)
        assert s_var is not None
        symbolic_vars.append(s_var)
    update_functions = {
        symbolic_vars[i]: graph.mk_update_function(network_vars[i])
        for i in range(len(network_vars))
    }

    # Use stochastic simulation to prune the set of candidate states.
    # We use different simulation approach depending on whether this space
    # is a minimal trap or not.
    if not avoid_bdd.is_false():
        candidates_bdd = state_list_to_bdd(symbolic_ctx, candidate_states)
        filtered_candidates: list[BooleanSpace] = []

        for i, state in enumerate(candidate_states):
            if i % 100 == 99 and sd.config["debug"]:
                print(
                    f"[{node_id}] > Simulation progress: {i + 1}/{len(candidate_states)}"
                )

            # Remove the state from the candidates. If we can prove that is
            # is not an attractor, we will put it back.
            state_bdd = graph.mk_subspace(state).to_bdd()
            candidates_bdd = candidates_bdd.l_and_not(state_bdd)

            # Convert the state BDD to a valuation object. This is like
            # a dictionary, but updates much faster because it is managed
            # by rust and indexed by symbolic variable IDs and not names.
            #
            # (The `state_bdd` should be a singleton, so it's ok to just
            # take the first valuation)
            simulation = state_bdd.valuation_first()

            is_valid_candidate = True
            for _ in range(max_iterations):
                # Advance all variables by one step in random order.
                generator.shuffle(symbolic_vars)
                for s_var in symbolic_vars:
                    # We can just "call" the BDD as a normal function with
                    # a valuation as the input.
                    step = update_functions[s_var](simulation)
                    simulation[s_var] = step

                # To check for set inclusion, we can also just "eval" the
                # BDD, since it evaluates to true for every state that is
                # in the encoded set.

                if candidates_bdd(simulation):
                    # The state can reach some other state in the candidate
                    # set. This does not mean it cannot be an attractor, but
                    # it means it is sufficient to keep considering
                    # the remaining candidates.
                    is_valid_candidate = False
                    break

                if avoid_bdd(simulation):
                    # The state can reach some other state in the avoid set,
                    # which means it cannot be an attractor in this subspace.
                    is_valid_candidate = False
                    break

            if is_valid_candidate:
                # If we cannot rule out the candidate, we have to put it back
                # into the candidate set.

                # However, we want to use the new state (after simulation),
                # because this one is probably "closer" to the goals that we
                # want to eventually get to.

                candidates_bdd = candidates_bdd.l_or(Bdd(simulation))
                filtered_candidates.append(valuation_to_state(symbolic_ctx, simulation))

        return filtered_candidates
    else:
        # If the avoid set is empty, it means that this is a pseudo-minimal space
        # and hence we (a) don't have to check if we reached the avoid set,
        # and (b) we can stop with one candidate instead of zero.

        candidates_bdd = state_list_to_bdd(symbolic_ctx, candidate_states)
        printed: set[int] = set()
        for i in range(max_iterations):
            progress = int((i * len(candidate_states)) / max_iterations)
            if (
                progress % 100 == 99
                and sd.config["debug"]
                and (progress not in printed)
            ):
                printed.add(progress)
                print(
                    f"[{node_id}] > Simulation progress: {progress + 1}/{len(candidate_states)}"
                )

            generator.shuffle(symbolic_vars)
            new_candidates_bdd = symbolic_ctx.mk_constant(False)

            # Goes through all the states in the candidates BDD, updates
            # them, and puts them in the new BDD assuming they are not there yet.
            for state_val in candidates_bdd.valuation_iterator():
                candidates_bdd = candidates_bdd.l_and_not(Bdd(state_val))

                simulation = state_val
                for s_var in symbolic_vars:
                    step = update_functions[s_var](simulation)
                    simulation[s_var] = step

                if candidates_bdd(simulation) or new_candidates_bdd(simulation):
                    # We have reached one of the other candidates.
                    continue

                new_candidates_bdd = new_candidates_bdd.l_or(Bdd(simulation))

            candidates_bdd = new_candidates_bdd

            if candidates_bdd.cardinality() <= 1:
                break

        # Finally, we convert the candidate states back into something that
        # biobalm understands.
        filtered_candidates = []
        for state_val in candidates_bdd.valuation_iterator():
            filtered_candidates.append(valuation_to_state(symbolic_ctx, state_val))

        return filtered_candidates


def asp_greedy_retained_set_optimization(
    sd: SuccessionDiagram,
    node_id: int,
    petri_net: DiGraph,
    retained_set: BooleanSpace,
    candidate_states: list[BooleanSpace],
    avoid_dnf: list[BooleanSpace],
) -> tuple[BooleanSpace, list[BooleanSpace]]:
    """
    Takes a Boolean network encoded as a Petri net and a candidate retained set.
    Then, performs a greedy optimization process leading to a locally optimal
    version of the original retained set that results in the least
    amount of candidate states.


    Parameters
    ----------
    node_id : int
        The ID of the associated SD node. This is only for logging progress.
    petri_net : networkx.DiGraph
        The Petri net encoding of the *percolated* network dynamics.
    retained_set : BooleanSpace
        The retained set that will be the initial point of the optimization.
    candidate_states: list[BooleanSpace]
        The list of candidates that are valid for the given `retained_set`, so
        that they don't need to be recomputed.
    avoid_dnf: list[BooleanSpace]
        The list of subspaces in the given network in which candidate
        states can be ignored.

    Returns
    -------
    tuple[BooleanSpace, list[BooleanSpace]]
        The optimized reatined set, together with the list of candidates that
        are valid for this retained set.
    """
    done = False
    while not done:
        done = True
        for var in retained_set:
            # Standrad termination checks.
            if len(candidate_states) == 0:
                return (retained_set, [])
            if len(avoid_dnf) == 0 and len(candidate_states) == 1:
                # If we are not avoiding anything, this is a (pseudo)
                # minimal space and a single candidate is good enough.
                return (retained_set, candidate_states)

            # Flip the value of the specified variable to see if the resulting
            # candidate set is smaller.
            retained_set_2 = retained_set.copy()
            retained_set_2[var] = cast(Literal[0, 1], 1 - retained_set_2[var])
            candidate_states_2 = compute_fixed_point_reduced_STG(
                petri_net,
                retained_set_2,
                avoid_subspaces=avoid_dnf,
                # We don't need all solutions if the result isn't smaller.
                solution_limit=len(candidate_states),
            )
            if len(candidate_states_2) < len(candidate_states):
                retained_set = retained_set_2
                candidate_states = candidate_states_2
                done = False
                if sd.config["debug"]:
                    print(
                        f"[{node_id}] > Candidate states optimized to {len(candidate_states)}."
                    )
    return (retained_set, candidate_states)


def make_heuristic_retained_set(
    graph: AsynchronousGraph, nfvs: list[str], avoid_dnf: list[BooleanSpace]
) -> BooleanSpace:
    """
    Calculate the retained set for a given `space` based on heuristic criteria.

    The retained set is technically a space-like object that describes the
    variables which have to be fixed in order for the network to lose all
    complex attractors. However, note that this really means changing the update
    functions of these variables. This is not a trap space that only contains
    fixed-points, but a description of how the network must be modified to
    eliminate all complex attractors.

    The construction guarantees that any complex attractor of the old
    network will manifest as at least one fixed-point in the new network. But this
    network modification is not implemented here. This method only generates
    the necessary list of variables and values.

    Finally, note that the method uses a heuristic to select values
    that should ideally lead to the least amount of fixed-points in
    the modified network, but there can be other, more optimal retained sets.

    Parameters
    ----------
    graph : AsynchronousGraph
        The symbolic update functions stored as an `AsynchronousGraph` object
        from the `biodivine_aeon` library.
    nfvs : list[str]
        The list of variables that represent an NFVS of the network encoded
        by `graph`.
    avoid_dnf : list[BooleanSpace]
        A list of :class:`BooleanSpace<biobalm.types.BooleanSpace>` objects
        describing the child spaces of the specified network. Only attractors that
        are not in these child spaces are considered. If no child spaces are
        provided, then all attractors are considered.

    Returns
    -------
    BooleanSpace
        A :class:`BooleanSpace<biobalm.types.BooleanSpace>` object describing the
        retained set.
    """
    retained_set: BooleanSpace = {}

    # First, if there are any child spaces present, we extend the retained set
    # with the values from the one that has the least amount of fixed variables
    # shared with the NFVS.
    if len(avoid_dnf) > 0:
        # Find the child space that has the fewest nodes in common with the NFVS:
        least_common_child_space = avoid_dnf[0]
        least_common_nodes = len(set(least_common_child_space) & set(nfvs))
        for child_space in avoid_dnf:
            common_nodes = len(set(child_space) & set(nfvs))
            if common_nodes < least_common_nodes:
                least_common_nodes = common_nodes
                least_common_child_space = child_space

        for x in least_common_child_space:
            if (x not in retained_set) and (x in nfvs):
                retained_set[x] = least_common_child_space[x]

    # Then, set the remaining NFVS variables based on the majority output value
    # in the update function of the relevant variable.
    for x in nfvs:
        if x in retained_set:
            continue

        fn_bdd = graph.mk_update_function(x)

        # If most of the function is positive, we set the value
        # to `1`, otherwise set it to `0`.
        if fn_bdd.cardinality() > fn_bdd.l_not().cardinality():
            retained_set[x] = 1
        else:
            retained_set[x] = 0

    return retained_set



from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from biobalm.succession_diagram import SuccessionDiagram


def expand_bfs(
    sd: SuccessionDiagram,
    node_id: int | None = None,
    bfs_level_limit: int | None = None,
    size_limit: int | None = None,
) -> bool:
    """
    See `SuccessionDiagram.expand_bfs` for documentation.
    """

    if node_id is None:
        node_id = sd.root()

    seen: set[int] = set()
    seen.add(node_id)

    level_id = 0
    current_level = [node_id]
    next_level: list[int] = []

    while len(current_level) > 0:
        for node in current_level:
            # Check if the size limit has been exceeded already.
            if (size_limit is not None) and (len(sd) >= size_limit):
                # Size limit reached.
                return False

            # Compute successors if necessary.
            successors = sd.node_successors(node, compute=True)
            # Sort successors to avoid non-determinism.
            successors = sorted(successors)

            # Add successors to the next level and to the seen set.
            for s in successors:
                if s not in seen:
                    seen.add(s)
                    next_level.append(s)

        # The level is explored. Check if this exceeds the level limit.
        if (bfs_level_limit is not None) and (level_id >= bfs_level_limit):
            # Level limit reached.
            return False

        # If not, "move on" to the next level.
        level_id += 1
        current_level = next_level
        next_level = []

    return True



"""
A module which provides utility methods for reachability analysis using Pint.

Note that in biobalm, we try to treat pint as an *optional* dependency. Hence, if you
want to use any of these methods, make sure to import the module in a way that fails
safely when Pint is not installed.
"""

from __future__ import annotations

from functools import reduce
from typing import TYPE_CHECKING, cast, Iterable

from networkx import DiGraph  # type: ignore

from biobalm.petri_net_translation import (
    place_to_variable,
    optimized_recursive_dnf_generator,
)
from biobalm.types import BooleanSpace, SuccessionDiagramConfiguration

if TYPE_CHECKING:
    from biodivine_aeon import Bdd


def pint_reachability(
    petri_net: DiGraph,
    initial_state: BooleanSpace,
    target_states: Bdd,
    config: SuccessionDiagramConfiguration,
) -> bool:
    """
    Use Pint to check if a given `initial_state` can possibly reach some state
    in the `target_states` BDD.

    If the reachability analysis is inconclusive, the method
    raises a `RuntimeError`.
    """
    from pypint import InMemoryModel  # type:ignore

    if target_states.is_false():
        return False  # Cannot reach a stat in an empty set.

    # Build a Pint model through an automata network and copy
    # over the initial condition.
    pint_model = InMemoryModel(_petri_net_as_automata_network(petri_net))
    for var, level in initial_state.items():
        pint_model.initial_state[var] = level

    goal = _pint_build_symbolic_goal(target_states, config)

    def failed(*_):
        raise RuntimeError("Cannot verify.")

    return pint_model.reachability(goal=goal, fallback=failed)  # type: ignore


def _pint_build_symbolic_goal(
    states: Bdd, config: SuccessionDiagramConfiguration
) -> object:
    """
    A helper method which (very explicitly) converts a set of states
    represented through a BDD into a Pint `Goal`.

    Note that if `GOAL_SIZE_LIMIT` is exceeded, a partial goal is returned
    that may not cover all the states in the argument `Bdd`. This avoids
    exceeding the argument list size limit, but introduces additional source
    of incompleteness into the reachability process.
    """
    from pypint import Goal  # type:ignore

    assert not states.is_false()

    goals: list[Goal] = []
    limit = config["pint_goal_size_limit"]
    for clause in optimized_recursive_dnf_generator(states):
        named_clause = {
            states.__ctx__().get_variable_name(var): int(value)
            for var, value in clause.items()
        }

        limit -= len(named_clause)
        if limit < 0:
            # If the goal is too large to be passed as a command line argument,
            # break here and don't continue. This is not ideal but I'm not sure
            # how to fix this other than modifying `pint` itself.
            if config["debug"]:
                print(
                    "WARNING: `pint` goal size limit exceeded. A partial goal is used."
                )
            break

        goal_atoms = [f"{var}={level}" for var, level in named_clause.items()]
        goals.append(Goal(",".join(goal_atoms)))

    return reduce(lambda a, b: a | b, goals)


def _petri_net_as_automata_network(petri_net: DiGraph) -> str:
    """
    Takes a Petri net which was created by implicant encoding from a Boolean network,
    and builds an automata network file (`.an`) compatible with the Pint tool.
    """
    automata_network = ""

    # Go through all PN places and save them as model variables.
    variable_set: set[str] = set()
    for place, kind in petri_net.nodes(data="kind"):  # type: ignore
        if kind != "place":
            continue
        variable_set.add(place_to_variable(place)[0])  # type: ignore[reportUnknownArgumentType] # noqa
    variables = sorted(variable_set)

    # Declare all variables with 0/1 domains.
    for var in variables:
        automata_network += f'"{var}" [0, 1]\n'

    for transition, kind in petri_net.nodes(data="kind"):  # type: ignore
        if kind != "transition":
            continue

        predecessors = set(cast(Iterable[str], petri_net.predecessors(transition)))  # type: ignore
        successors = set(cast(Iterable[str], petri_net.successors(transition)))  # type: ignore

        # The value under modification is the only
        # value that is different between successors and predecessors.
        source_place = next(iter(predecessors - successors))
        target_place = next(iter(successors - predecessors))

        (s_var, s_level) = place_to_variable(source_place)
        (t_var, t_level) = place_to_variable(target_place)
        assert s_var == t_var

        # The remaining places represent the necessary conditions.
        # Here, we transform them into a text format.
        condition_places = sorted(predecessors.intersection(successors))
        condition_tuples = [place_to_variable(p) for p in condition_places]
        conditions = [f'"{var}"={int(level)}' for var, level in condition_tuples]

        # A pint rule consists of a variable name, value transition,
        # and a list of necessary conditions for the transition (if any).
        if len(conditions) == 0:
            rule = f'"{s_var}" {int(s_level)} -> {int(t_level)}\n'
        else:
            rule = f"\"{s_var}\" {int(s_level)} -> {int(t_level)} when {' and '.join(conditions)}\n"

        automata_network += rule

    return automata_network



"""
Here, we implement the Trappist method for computing fixed-points, minimum trap spaces
and maximum trap spaces, including time-reversed networks.

The implementation is based on the official [trappist package](https://github.com/soli/trap-spaces-as-siphons),
but has been significantly modified to better suite this project.
"""

from __future__ import annotations

from typing import TYPE_CHECKING, Literal

if TYPE_CHECKING:
    from typing import Callable

    from clingo import Model

    from biobalm.types import BooleanSpace

from biodivine_aeon import BooleanNetwork
from clingo import Control, SolveHandle
from networkx import DiGraph  # type: ignore

from biobalm.petri_net_translation import (
    extract_source_variables,
    extract_variable_names,
    network_to_petrinet,
    place_to_variable,
    variable_to_place,
)


def trappist_async(
    network: BooleanNetwork | DiGraph,
    on_solution: Callable[[BooleanSpace], bool],
    problem: Literal["min", "max", "fix"] = "min",
    reverse_time: bool = False,
    ensure_subspace: BooleanSpace | None = None,
    avoid_subspaces: list[BooleanSpace] | None = None,
    optimize_source_variables: list[str] | None = None,
):
    """
    Asynchronous version of the :func:`trappist` method.

    *Note that "asynchronous" refers to the execution of this function, not to
    the update scheme used (which does not affect the trap spaces in any case).*

    The same as the :func:`trappist` method, but instead of returning a list of spaces
    as a result, the spaces are returned to the supplied `on_solution` callback.
    You can stop the enumeration by returning `False` from this callback.

    See :func:`trappist` for details.
    """
    if ensure_subspace is None:
        ensure_subspace = {}
    if avoid_subspaces is None:
        avoid_subspaces = []

    if isinstance(network, BooleanNetwork):
        bn = network
        petri_net = network_to_petrinet(network)
    else:
        bn = None
        petri_net = network

    assert isinstance(petri_net, DiGraph)

    if bn is None:
        variables = extract_variable_names(petri_net)
    else:
        variables = [bn.get_variable_name(v) for v in bn.variables()]

    if optimize_source_variables is None:
        optimize_source_variables = extract_source_variables(petri_net)

    ctl = _create_clingo_constraints(
        variables,
        petri_net,
        problem,
        reverse_time,
        ensure_subspace,
        avoid_subspaces,
        optimize_source_variables,
    )

    ctl.ground()
    result = ctl.solve(yield_=True)
    if isinstance(result, SolveHandle):
        with result as iterator:
            for model in iterator:
                if not on_solution(_clingo_model_to_space(model)):
                    break
    # Else: unsat, hence we don't do anything.


def trappist(
    network: BooleanNetwork | DiGraph,
    problem: Literal["min", "max", "fix"] = "min",
    reverse_time: bool = False,
    solution_limit: int | None = None,
    ensure_subspace: BooleanSpace | None = None,
    avoid_subspaces: list[BooleanSpace] | None = None,
    optimize_source_variables: list[str] | None = None,
) -> list[BooleanSpace]:
    """
    Trap space solver for Boolean networks.

    Solves the given `problem` for the given `network` using the Trappist
    algorithm, internally relying on the Python bindings of the `clingo` ASP
    solver.

    The `problem` can be either `"min"`, `"max"` or `"fix"`, corresponding to
    finding all minimal trap spaces (`"min"`), maximal trap spaces (`"max"`), or
    fixed points (`"fix"`).

    The result is a list of spaces represented as :class:`BooleanSpace<biobalm.types.BooleanSpace>`
    dictionaries. If you want to avoid enumerating all solutions explicitly as one list, you can use
    :func:`trappist_async` which has a similar API but can yield solutions one by one.

    Finally, recall that the supplied network must have its names sanitized (see
    :mod:`petri_net_translation<biobalm.petri_net_translation>` module).


    Parameters
    ----------
    network : BooleanNetwork | DiGraph
        A Boolean network or a Petri net compatible with the encoding in
        :mod:`petri_net_translation<biobalm.petri_net_translation>` module. The behaviour is undefined for other
        `DiGraph` instances.
    problem : Literal["min", "max", "fix"], optional
        The problem to solve. Finds all minimal trap spaces (`"min"`), maximal
        trap spaces (`"max"`), or fixed points (`"fix"`). The default is `"min"`.
    reverse_time : bool, optional
        If `True`, a time-reversed network should be considered. The default is
        `False`.
    solution_limit : int, optional
        If given, the result is limited to the given number of solutions.
        Default: `None`.
    ensure_subspace : BooleanSpace, optional
        If given, the result is restricted to the given subspace. Default: `None`.
    avoid_subspaces : list[BooleanSpace], optional
        If given, the result is restricted to not be in the given subspaces. Default:
        `None`.
    optimize_source_variables : list[str], optional
        Desingates which variables should be treated as input nodes for the
        purposes of trap space identification. Fixed values of these variables
        will be considered together, reducing the number of trap spaces
        associated with `k` source nodes to `2**k` from `(2**k)*k!`.

    Returns
    -------
    list[BooleanSpace]
        The :class:`BooleanSpace<biobalm.types.BooleanSpace>` objects
        describing the trap spaces that solve the specified problem.
    """
    if ensure_subspace is None:
        ensure_subspace = {}
    if avoid_subspaces is None:
        avoid_subspaces = []

    results: list[BooleanSpace] = []

    def save_result(x: BooleanSpace) -> bool:
        results.append(x)
        if solution_limit is None:
            return True
        else:
            return len(results) < solution_limit

    trappist_async(
        network,
        on_solution=save_result,
        problem=problem,
        reverse_time=reverse_time,
        ensure_subspace=ensure_subspace,
        avoid_subspaces=avoid_subspaces,
        optimize_source_variables=optimize_source_variables,
    )

    return results


def _clingo_model_to_space(model: Model) -> BooleanSpace:
    space: BooleanSpace = {}
    for atom in model.symbols(atoms=True):
        atom_str = str(atom)
        (variable, is_positive) = place_to_variable(atom_str)
        # This should be prevented by the "conflic-free" property of the result,
        # but just in case.
        assert variable not in space
        # Note that this is counterintuitive but correct. If "positive" symbol
        # appears in the solution, we want to fix the value to 0. This is indeed
        # the intended behaviour of the algorithm.
        space[variable] = 0 if is_positive else 1
        space[variable] = 0 if is_positive else 1
    return space


def _create_clingo_constraints(
    variables: list[str],
    petri_net: DiGraph,
    problem: str = "min",
    reverse_time: bool = False,
    ensure_subspace: BooleanSpace | None = None,
    avoid_subspaces: list[BooleanSpace] | None = None,
    optimize_source_variables: list[str] | None = None,
) -> Control:
    """
    Translate the given Petri net into a logic program that solves
    the given problem type.

    All parameters are as specified in :func:`trappist`.
    """
    if ensure_subspace is None:
        ensure_subspace = {}
    if avoid_subspaces is None:
        avoid_subspaces = []
    if optimize_source_variables is None:
        optimize_source_variables = []

    assert problem in ["min", "max", "fix"], f"Unknown problem type: {problem}"

    dom_mod = "--dom-mod=3, 16"  # for min. trap spaces and fixed points
    if problem == "max":
        dom_mod = "--dom-mod=5, 16"  # for max. trap spaces

    # "0" specifies that all solutions should be listed (we implement the limit
    # later using callbacks).
    # TODO: Explain what remaining options mean and why we need them?
    ctl = Control(["0", "--heuristic=Domain", "--enum-mod=domRec", dom_mod])

    # Declare places and their conflicts based on network variables.
    for var_name in variables:
        p_name = variable_to_place(var_name, positive=True)
        n_name = variable_to_place(var_name, positive=False)
        # Declare a positive and negative symbol.
        ctl.add(f"{{{p_name}}}.")
        ctl.add(f"{{{n_name}}}.")
        # Assert there is no conflict (we can't have both true simultaneously).
        ctl.add(f":- {p_name}, {n_name}.")
        # If we are computing fixed points, assert also that at least one should hold.
        if problem == "fix":
            ctl.add(f"{p_name} ; {n_name}.")

    # Ensure that solutions must have desired variables fixed based on `ensure_subspace`.
    for fixed_var in ensure_subspace:
        positive = True
        if ensure_subspace[fixed_var] == 1:
            positive = False
        ctl.add(f"{variable_to_place(fixed_var, positive)}.")

    # Ensure that solutions can't have variables fixed based on either subspace in `avoid_subspaces`.
    for to_avoid in avoid_subspaces:
        fixed_list = [variable_to_place(var, (to_avoid[var] != 1)) for var in to_avoid]
        fixed_vars = ", ".join(fixed_list)
        ctl.add(f":- {fixed_vars}.")

    free_places: list[str] = []
    for node, kind in petri_net.nodes(data="kind"):  # type: ignore # noqa
        assert isinstance(node, str)
        if kind == "place":
            if place_to_variable(node)[0] not in ensure_subspace:
                free_places.append(node)
        elif kind == "transition":
            if not reverse_time:
                # Compute siphons.
                predecessors: list[str] = list(petri_net.predecessors(node))  # type: ignore # noqa
                p_disjunction = "; ".join(predecessors)
                for successor in petri_net.successors(node):  # type: ignore # noqa
                    if (
                        successor not in predecessors
                    ):  # optimize obvious tautologies # noqa
                        ctl.add(f"{p_disjunction} :- {successor}.")
            else:
                # Compute traps.
                successors = list(petri_net.successors(node))  # type: ignore # noqa
                s_disjunction = "; ".join(successors)  # type: ignore # noqa
                for predecessor in petri_net.predecessors(node):  # type: ignore # noqa
                    if predecessor not in successors:
                        ctl.add(f"{s_disjunction} :- {predecessor}.")
        else:
            raise Exception(f"Unexpected node kind: `{kind}`.")

    # For maximal trap spaces, we need an extra condition.
    if problem == "max" and len(free_places) > 0:
        # Only spaces which are not fixed but the `ensure_subspace` are considered here.
        max_condition = "; ".join(free_places)
        ctl.add(f"{max_condition}.")

        # Additionally, we require source nodes to not appear as `*` when `problem=max`, since this is
        # largely useless in practice.
        for variable in optimize_source_variables:
            if variable not in ensure_subspace:
                ctl.add(
                    f"{variable_to_place(variable, True)}; {variable_to_place(variable, False)}."
                )

    return ctl


def _clingo_model_to_fixed_point(model: Model) -> BooleanSpace:
    """
    Convert a clingo `Model` to a subspace representing a single fixed point.
    That is, the space should have all model variables fixed.

    This is a "positive" conversion, because the method for computing fixed points
    produces "positive" models (i.e. the space is represented by positive atoms
    present in the model).
    """
    space: BooleanSpace = {}

    for atom in model.symbols(atoms=True):
        atom_str = str(atom)
        (variable, is_positive) = place_to_variable(atom_str)

        # This should be prevented by the "conflic-free" property of the result,
        # but just in case.
        assert variable not in space

        # Note that this is opposite to the case of trap spaces. If "positive"
        # symbol appears in the solution, we fix the value to "1". Otherwise, we
        # fix the value to "0".

        space[variable] = 1 if is_positive else 0

    return space


def _create_clingo_fixed_point_constraints(
    variables: list[str],
    petri_net: DiGraph,
    ensure_subspace: BooleanSpace | None = None,
    avoid_subspaces: list[BooleanSpace] | None = None,
) -> Control:
    """
    Generate the ASP characterizing all deadlocks of the Petri net (equivalently all
    fixed points of the Boolean network).
    """
    if ensure_subspace is None:
        ensure_subspace = {}
    if avoid_subspaces is None:
        avoid_subspaces = []
    dom_mod = "--dom-mod=3, 16"  # for fixed points

    # "0" specifies that all solutions should be listed (we implement the limit
    # later using callbacks).
    # TODO: Explain what remaining options mean and why we need them?
    ctl = Control(["0", "--heuristic=Domain", "--enum-mod=domRec", dom_mod])

    # Declare places and their conflicts based on network variables.
    for node in variables:
        p_name = variable_to_place(node, positive=True)
        n_name = variable_to_place(node, positive=False)

        # Declare a positive and negative symbol.
        ctl.add("base", [], f"{{{p_name}}}.")
        ctl.add("base", [], f"{{{n_name}}}.")

        # Assert there is a fixed point.
        ctl.add("base", [], f":- {p_name}, {n_name}.")
        ctl.add("base", [], f"{p_name} ; {n_name}.")

    for node, kind in petri_net.nodes(data="kind"):  # type: ignore
        if kind == "place":
            continue
        elif kind == "transition":
            preds = list(petri_net.predecessors(node))  # type: ignore

            pred_rhs = "; ".join(preds)  # type: ignore
            ctl.add("base", [], f":- {pred_rhs}.")
        else:
            raise Exception(f"Unexpected node kind: `{kind}`.")

    # Ensure that solutions must have desired variables fixed based on `ensure_subspace`.
    for fixed_var, value in ensure_subspace.items():
        place_name = variable_to_place(fixed_var, positive=bool(value))
        ctl.add("base", [], f"{place_name}.")

    # Ensure that fixed points can't lie in either subspace in `avoid_subspaces`.
    for to_avoid in avoid_subspaces:
        if len(to_avoid) > 0:
            """
            Note that this is opposite to the case of trap spaces.
            m(x) = 0 ~ place b0_x and m(x) = 1 ~ place b1_x
            """
            fixed_list = [
                variable_to_place(var, (to_avoid[var] == 1)) for var in to_avoid
            ]
            fixed_vars = ", ".join(fixed_list)
            ctl.add("base", [], f":- {fixed_vars}.")
        else:
            ctl.add("base", [], "#false.")
            break  # There is no solution and we do not need to process more.

    return ctl


def compute_fixed_point_reduced_STG_async(
    petri_net: DiGraph,
    retained_set: BooleanSpace,
    on_solution: Callable[[BooleanSpace], bool],
    ensure_subspace: BooleanSpace | None = None,
    avoid_subspaces: list[BooleanSpace] | None = None,
):
    """
    Asynchronous version of :func:`compute_fixed_point_reduced_STG`.

    *Note that "asynchronous" refers to the execution of this function, not to
    the update scheme used to generate the STG (which is, coincidently, always
    asynchronous).*

    The same as the :func:`compute_fixed_point_reduced_STG`, but instead of returning a
    list of fixed-points as a result, the states are returned to the supplied
    `on_solution` callback. You can stop the enumeration by
    returning `False` from this callback.

    See :func:`compute_fixed_point_reduced_STG` for details.
    """
    if ensure_subspace is None:
        ensure_subspace = {}
    if avoid_subspaces is None:
        avoid_subspaces = []

    # Build a copy of the original Petri net where the
    # variables in the retained set can only change their
    # value towards the "retain value".
    reduced_petri_net: DiGraph = petri_net.copy()  # type: ignore
    for node in retained_set.keys():
        b_i = retained_set[node]
        source_place = variable_to_place(node, positive=(b_i == 1))

        preds = list(reduced_petri_net.predecessors(source_place))  # type: ignore # noqa
        succs = list(reduced_petri_net.successors(source_place))  # type: ignore # noqa

        deleted_transitions = list(set(succs) - set(preds))  # type: ignore # noqa

        for trans in deleted_transitions:  # type: ignore # noqa
            reduced_petri_net.remove_node(trans)  # type: ignore # noqa

    ctl = _create_clingo_fixed_point_constraints(
        extract_variable_names(reduced_petri_net),
        reduced_petri_net,
        ensure_subspace,
        avoid_subspaces,
    )

    ctl.ground([("base", [])])
    result = ctl.solve(yield_=True)
    if isinstance(result, SolveHandle):
        with result as iterator:
            for model in iterator:
                if not on_solution(_clingo_model_to_fixed_point(model)):
                    break
    # Else: unsat, hence we don't do anything.


def compute_fixed_point_reduced_STG(
    petri_net: DiGraph,
    retained_set: BooleanSpace = {},
    ensure_subspace: BooleanSpace = {},
    avoid_subspaces: list[BooleanSpace] = [],
    solution_limit: int | None = None,
) -> list[BooleanSpace]:
    """
    Compute fixed points of the given Petri-net-encoded Boolean network.

    This method computes the fixed points of the given Petri-net-encoded Boolean
    network faster than using the :func:`trappist` function (which considers
    all minimal trap spaces, not just fixed points).

    If you want to avoid enumerating all solutions explicitly as one list, you can use
    :func:`trappist_async` which has a similar API but can yield solutions one by one.

    Parameters
    ----------
    petri_net : DiGraph
        The Petri net which was created by the implicant encoding from a
        Boolean network. See :mod:`petri_net_translation<biobalm.petri_net_translation>`
        for details.
    retained_set : BooleanSpace
        If set, specifies a set of variables that should only be allowed to update
        towards their "retained value". This condition is applied by modifying
        the structure of the Petri net encoding.
    ensure_subspace : BooleanSpace
        Only fixed points in this subspace will be considered or returned.
    avoid_subspaces : list[BooleanSpace]
        Only fixed points not in any of these subspaces will be considered or
        returned.
    solution_limit : int
        The maximum number of solutions to return. If `None`, all solutions will
        be returned.

    Returns
    -------
    list[BooleanSpace]
        A list of fixed points.
    """

    results: list[BooleanSpace] = []

    def save_result(x: BooleanSpace) -> bool:
        results.append(x)
        if solution_limit is None:
            return True
        else:
            return len(results) < solution_limit

    compute_fixed_point_reduced_STG_async(
        petri_net,
        retained_set,
        on_solution=save_result,
        ensure_subspace=ensure_subspace,
        avoid_subspaces=avoid_subspaces,
    )
    return results



from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from biobalm.succession_diagram import SuccessionDiagram

from biobalm.types import BooleanSpace
from biobalm.space_utils import intersect
from biobalm.trappist_core import compute_fixed_point_reduced_STG
from biobalm._sd_attractors.attractor_candidates import make_heuristic_retained_set
from biodivine_aeon import AsynchronousGraph


def expand_attractor_seeds(sd: SuccessionDiagram, size_limit: int | None = None):
    """
    See `SuccessionDiagram.expand_attractor_seeds` for documentation.
    """

    # First, expand the succession diagram such that all minimal trap spaces are
    # found. This reduces the amount of work performed in this algorithm,
    # because for every attractor in a minimal trap space, we already have the
    # closest trap space, now we just need to do the same for (potential)
    # motif-avoidant attractors.
    sd.expand_minimal_spaces(size_limit=size_limit)

    if sd.config["debug"]:
        print(
            "Minimal trap space expansion finished. Proceeding to attractor expansion."
        )

    root = sd.root()
    seen = set([root])
    stack: list[tuple[int, list[int] | None]] = [(root, None)]

    while len(stack) > 0:
        (node, successors) = stack.pop()
        if successors is None:
            # Only allow successor computation if size limit hasn't been exceeded.
            if (size_limit is not None) and (len(sd) >= size_limit):
                # Size limit reached.
                return False

            successors = sd.node_successors(node, compute=True)
            successors = sorted(successors, reverse=True)  # For determinism!
            # (reversed because we explore the list from the back)

        # Retrieve the stable motifs of children that are already expanded.
        expanded_children = [
            x for x in sd.node_successors(node) if sd.node_data(x)["expanded"]
        ]
        expanded_motifs = [
            sd.edge_stable_motif(node, child) for child in expanded_children
        ]

        # Now, we skip all successors that are either already seen, or that
        # do not contain any candidate states for motif-avoidant attractors.
        while len(successors) > 0:
            if successors[-1] in seen:
                # The next node was already seen on stack. We can thus skip it
                # and continue to the next one.
                successors.pop()
                continue
            if sd.node_data(successors[-1])["expanded"]:
                # The next node to explore is expanded (by some previous procedure)
                # but not "seen" in this search yet. We need to visit this node
                # regardless of other conditions
                break
            # Now, we need to asses if the next successor has some candidate states which
            # are not covered by the already expanded children.

            s = successors[-1]

            successor_space = sd.node_data(s)["space"]
            successor_bn = sd.node_percolated_network(s, compute=True)
            successor_nfvs = sd.node_percolated_nfvs(s, compute=True)
            successor_pn = sd.node_percolated_petri_net(s, compute=True)
            successor_graph = AsynchronousGraph(successor_bn)

            avoid_or_none = [
                intersect(successor_space, child) for child in expanded_motifs
            ]
            avoid = [x for x in avoid_or_none if x is not None]
            avoid_restricted: list[BooleanSpace] = []
            for x in avoid:
                y: BooleanSpace = {
                    var: val for (var, val) in x.items() if var not in successor_space
                }
                avoid_restricted.append(y)

            retained_set = make_heuristic_retained_set(
                successor_graph, successor_nfvs, avoid
            )

            successor_seeds = compute_fixed_point_reduced_STG(
                successor_pn,
                retained_set,
                avoid_subspaces=avoid_restricted,
                solution_limit=1,
            )

            if len(successor_seeds) == 0:
                # At this point, we know that this successor is not expanded and
                # there are either no candidate states in it, or all candidate
                # states are already covered by some other expanded successor.
                successors.pop()
                continue

            if sd.config["debug"]:
                print(
                    f"[{node}] Found successor with new attractor candidate seeds. Expand node {successors[-1]}."
                )

            break

        if len(successors) == 0:
            # Everything is done for this `node` and we can continue to the next one.
            if sd.config["debug"]:
                print(f"[{node}] Finished node attractor expansion.")
            continue

        s = successors.pop()
        seen.add(s)
        # Push the node back with the remaining successors.
        stack.append((node, successors))
        # Push the successor onto the stack.
        stack.append((s, None))

    return True



"""
A utility module which provides several methods for manipulating the influence graph
of a Boolean network.
"""

from __future__ import annotations

from typing import TYPE_CHECKING, Literal

if TYPE_CHECKING:
    from typing import Any, Sequence

    from biodivine_aeon import Regulation, VariableId

from typing import cast

from biodivine_aeon import BooleanNetwork, RegulatoryGraph, SignType, SymbolicContext
from networkx import DiGraph  # type: ignore


def _digraph_to_regulatory_graph(graph: DiGraph) -> RegulatoryGraph:
    """
    A helper method to convert a `networkx.DiGraph` to a `biodivine_aeon.RegulatoryGraph`.

    All nodes of the digraph should have string identifiers. Edges can be optionally
    annotated with a `sign` value `"+"`, `"-"` or `"?"` (however, `"?"` is treated the
    same as a missing `sign` annotation).

    Parameters
    ----------
    graph : DiGraph
        A `networkx.DiGraph` with string identifiers as nodes and optional
        `sign` annotations for edges.

    Returns
    -------
    RegulatoryGraph
        A `biodivine_aeon.RegulatoryGraph` with the same nodes and edges as `graph`.
    """

    rg = RegulatoryGraph(list(graph.nodes()))  # type: ignore
    for edge in graph.edges():  # type: ignore
        edge_data: dict[Any, Any] = graph.get_edge_data(edge[0], edge[1])  # type: ignore
        reg: Regulation = {
            "source": edge[0],
            "target": edge[1],
            "essential": True,
            "sign": None,
        }
        if "sign" in edge_data and edge_data["sign"] in ["+", "-"]:
            reg["sign"] = cast(SignType, edge_data["sign"])

        rg.add_regulation(reg)

    return rg


def feedback_vertex_set(
    network: RegulatoryGraph | DiGraph,
    parity: Literal["positive", "negative"] | None = None,
    subgraph: Sequence[str | VariableId] | None = None,
) -> list[str]:
    """
    Compute an approximately minimal feedback vertex set (FVS).

    Uses the implementation of `biodivine_aeon`.

    An FVS is a set of nodes in a network whose removal from the network
    results in an acyclic network. An FVS is minimal if it does not contain any
    FVS as a proper subset. This method uses a heuristic approach to attempt to find a
    minimal FVS. The result is guaranteed to be a feedback vertex set, but it might not be minimal.

    A paritiy can be specified. If the parity is specified, only cycles of the
    specified parity are considered (e.g. if `parity='negative'`, there can still be positive
    cycles in the graph not covered by the returned feedback vertex set).

    The method is deterministic (the same pseudo-optimal FVS is returned every time).

    Parameters
    ----------
    network : RegulatoryGraph | DiGraph
        Network to consider. If a `DiGraph` is given, it is converted to a `RegulatoryGraph`.
    parity : Literal["positive";, "negative"] | None, optional
        Must be `"positive"`, `"negative"`, or `None` (default). If
        `"positive"`, only cycles with an even number of negative edges are
        considered. If `"negative"`, only cycles with an odd number of negative
        edges are considered. Edges with no monotonicity are counted as both
        positive and negative. Otherwise, all cycles are considered.
    subgraph : Sequence[str  |  VariableId] | None, optional
        A list of network variables (either string names or AEON `VariableId`
        objects are fine). If given, the result is the FVS of the sub-graph induced by
        these network nodes.

    Returns
    -------
    list[str]
        A list of node names representing the smallest FVS found during the
        search. Sorted in the same order as in the input network (usually
        lexicographically).

    Examples
    --------
    >>> import biobalm
    >>> from biobalm.interaction_graph_utils import feedback_vertex_set
    >>> sd = biobalm.SuccessionDiagram.from_rules(\"\"\"
    ...     A, B
    ...     B, A
    ...     C, D
    ...     D, !C\"\"\")
    >>> feedback_vertex_set(sd.network)
    ['A', 'C']
    >>> feedback_vertex_set(sd.network, parity="positive")
    ['A']
    >>> feedback_vertex_set(sd.network, parity="negative")
    ['C']
    """
    if isinstance(network, DiGraph):
        network = _digraph_to_regulatory_graph(network)
    assert isinstance(network, RegulatoryGraph)
    fvs = network.feedback_vertex_set(parity, subgraph)
    return sorted([network.get_variable_name(x) for x in fvs])


def cleanup_network(network: BooleanNetwork) -> BooleanNetwork:
    """
    Prepare a `BooleanNetwork` object for use in a `SuccessionDiagram`. This
    mainly ensures that the network does not use parameters and fixes all static
    constraints to ensure that they are actually correct.

    Parameters
    ----------
    network : BooleanNetwork
        The Boolean network to be prepared.

    Returns
    -------
    BooleanNetwork
        The prepared network.
    """

    assert (
        network.explicit_parameter_count() == 0
    ), f"Parametrized networks are not supported. Found parameters: {network.explicit_parameter_names()}."

    # Implicit parameters with no regulators are allowed, since they just reprtesent free inputs
    # and are explicitly handled by the succession diagram.
    non_input_implicit = [
        v for v in network.implicit_parameters() if len(network.predecessors(v)) > 0
    ]
    if len(non_input_implicit) > 0:
        names = [network.get_variable_name(x) for x in non_input_implicit]
        raise AssertionError(
            f"Parametrized networks are not supported. Found implicit parameters: {names}."
        )

    return network.infer_valid_graph()


def source_SCCs(bn: BooleanNetwork) -> list[list[str]]:
    """
    Find source SCCs of the given `BooleanNetwork`.

    Here, SCC stands for "strongly connected component". An SCC is a source SCC
    if it has no incoming edges.

    Parameters
    ----------
    bn : BooleanNetwork
        The Boolean network to be examined.

    Returns
    -------
    list[list[str]]
        The list of source SCCs.
    """
    result: list[list[str]] = []
    for scc in bn.strongly_connected_components():
        scc_list = sorted(scc)
        if bn.backward_reachable(scc_list) == scc:
            scc_names = [bn.get_variable_name(var) for var in scc_list]
            result.append(scc_names)

    return sorted(result)


def source_nodes(
    network: BooleanNetwork, ctx: SymbolicContext | None = None
) -> list[str]:
    """
    Identify the source nodes of a given `BooleanNetwork`.

    Return the source nodes of a `BooleanNetwork`. That is, variables whose
    value cannot change, but is not fixed to a `true`/`false` constant.

    Note that this internally uses BDD translation to detect identity functions
    semantically rather than syntactically. If you already have a
    `SymbolicContext` for the given `network` available, you can supply it as
    the second argument.

    Parameters
    ----------
    network : BooleanNetwork
        The Boolean network to be examined.
    ctx : SymbolicContext
        The context used to translate the network to BDDs. A
        `biodivine_aeon.SymbolicContext` object.

    Returns
    -------
    list[str]
        The list of source nodes.
    """
    if ctx is None:
        ctx = SymbolicContext(network)

    result: list[str] = []
    for var in network.variable_names():
        update_function = network.get_update_function(var)
        if update_function is None:
            # This is an input variable with unspecified update
            # (this defaults to identity in most tools).
            assert len(network.predecessors(var)) == 0
            result.append(var)
        else:
            fn_bdd = ctx.mk_update_function(update_function)
            var_bdd = ctx.mk_network_variable(var)
            if fn_bdd == var_bdd:
                result.append(network.get_variable_name(var))

    return result



"""
This is an internal module which contains the implementations of the more
involved algorithms related to `SuccessionDiagram` construction and
manipulation. The idea is that any implementation that requires some helper
methods or data structures should be present here instead of
`SuccessionDiagram.py` to avoid mixing of code used by different algorithms and
bloating the core succession diagram implementation.

The algorithms within this module here should only use the public API of the
`SuccessionDiagram` to avoid violating its invariants. The actual algorithms are
then "re-exported" as a public methods of the `SuccessionDiagram` class.
"""



from __future__ import annotations

from typing import TYPE_CHECKING, cast, Literal

if TYPE_CHECKING:
    from biobalm.succession_diagram import SuccessionDiagram
    from biodivine_aeon import VariableId

from biodivine_aeon import (
    AsynchronousGraph,
    ColoredVertexSet,
    VertexSet,
    Attractors,
    Reachability,
)
from biobalm.symbolic_utils import state_list_to_bdd
from biobalm.space_utils import is_subspace, intersect
from biobalm.types import BooleanSpace
import biodivine_aeon
import copy


def symbolic_attractor_fallback(
    sd: SuccessionDiagram,
    node_id: int,
) -> tuple[list[BooleanSpace], list[VertexSet]]:
    """
    In case the attractor candidates cannot be computed, this fallback method
    can be used to attempt fully symbolic attractor computation for the given node.

    If the normal method works, then it is usually faster, but in the rare cases
    where it fails, this can sometimes help to solve the few outlier nodes that
    would otherwise block the computation.
    """

    old_log_level = biodivine_aeon.LOG_LEVEL

    if sd.config["debug"]:
        print(f"[{node_id}] > Start symbolic fallback.")
        biodivine_aeon.LOG_LEVEL = biodivine_aeon.LOG_ESSENTIAL

    node_data = sd.node_data(node_id)
    node_space = node_data["space"]

    candidates = sd.symbolic.mk_subspace(node_space)
    if node_data["expanded"]:
        # This node has successors that we should exclude from the attractor search.
        for s in sd.node_successors(node_id, compute=False):
            s_space = sd.node_data(s)["space"]
            candidates = candidates.minus(sd.symbolic.mk_subspace(s_space))

    if node_data["skipped"]:
        # This is the same method that we applied to candidate states computation.
        initial_size = candidates.cardinality()
        for n in sd.node_ids():
            n_data = sd.node_data(n)
            if is_subspace(node_space, n_data["space"]):
                continue
            if n_data["attractor_candidates"] == [] or n_data["attractor_seeds"] == []:
                # This will create a lot of duplicates, but it seems to be better than
                # not doing it at all.
                common_subspace = intersect(node_space, n_data["space"])
                if common_subspace is not None:
                    candidates = candidates.minus(
                        sd.symbolic.mk_subspace(common_subspace)
                    )
        if sd.config["debug"]:
            print(
                f"[{node_id}] Simplified symbolic fallback candidates with skip nodes from {initial_size} to {candidates.cardinality()}."
            )

    # These should cover all cycles in the network, so transition-guided reduction
    # only needs to consider these variables.
    internal_nfvs = sd.node_percolated_nfvs(node_id, compute=True)

    if sd.config["debug"]:
        print(f"[{node_id}] > Initial attractor candidates: {candidates}")

    candidates = Attractors.transition_guided_reduction(
        sd.symbolic,
        candidates,
        internal_nfvs,
    )

    if not sd.node_is_minimal(node_id) and not candidates.is_empty():
        # This like it could be a motif-avoidant attractor. We better investigate this further,
        # because this could get complicated...
        avoid = sd.symbolic.mk_empty_colored_vertices()
        if node_data["expanded"]:
            for s in sd.node_successors(node_id):
                s_space = sd.node_data(s)["space"]
                avoid = avoid.union(sd.symbolic.mk_subspace(s_space))

        if sd.config["debug"]:
            print(
                f"[{node_id}] > Reduction was ineffective. Start computing avoid set: {avoid}"
            )

        avoid = Reachability.reach_bwd(sd.symbolic, avoid)

        if sd.config["debug"]:
            print(f"[{node_id}] > Avoid set: {avoid}")

        candidates = candidates.minus(avoid)
    else:
        if sd.config["debug"]:
            print(
                f"[{node_id}] > Node is minimal and the reduction did not finish with an empty set. This is fine."
            )

    if sd.config["debug"]:
        print(f"[{node_id}] > Attractor candidates after reduction: {candidates}")

    attractors = Attractors.xie_beerel(sd.symbolic, candidates)

    biodivine_aeon.LOG_LEVEL = old_log_level

    result_seeds: list[BooleanSpace] = []
    result_sets: list[VertexSet] = []
    for attr in attractors:
        attr_vertices = attr.vertices()
        attr_seed = next(attr_vertices.items()).to_dict()
        attr_seed_named = {
            sd.network.get_variable_name(k): v for (k, v) in attr_seed.items()
        }
        result_seeds.append(cast(BooleanSpace, attr_seed_named))
        result_sets.append(attr_vertices)

    return (result_seeds, result_sets)


def compute_attractors_symbolic(
    sd: SuccessionDiagram,
    node_id: int,
    candidate_states: list[BooleanSpace],
    seeds_only: bool = False,
) -> tuple[list[BooleanSpace], list[VertexSet] | None]:
    """
    Uses exhaustive symbolic reachability to eliminate spurious candidate states
    and to compute the exact attractor sets.

    Parameters
    ----------
    sd : SuccessionDiagram
        The succession diagram in question.
    node_id : int
        An ID of an SD node for which we are computing the candidates.
    candidate_states: bool
        The list of candidate states covering the attractors in the
        given node.
    seeds_only: bool
        If `True`, the method can terminate early once it is guaranteed that
        all seeds have been identified. In such case, the list of returned sets
        is `None`.

    Returns
    -------
    tuple[list[BooleanSpace], list[VertexSet] | None]
        The list of attractor seed states and corresponding attractor sets (if computed).
    """

    node_data = sd.node_data(node_id)
    node_space = node_data["space"]
    bn_reduced = sd.node_percolated_network(node_id, compute=True)
    graph_reduced = AsynchronousGraph(bn_reduced)
    symbolic_ctx = graph_reduced.symbolic_context()

    # Ignore variables that are already fixed in the nodes space.
    # These are not used in the reduced network.
    candidate_states_reduced: list[BooleanSpace] = []
    for candidate in candidate_states:
        candidate_reduced: BooleanSpace = {
            k: v for (k, v) in candidate.items() if k not in node_space
        }
        candidate_states_reduced.append(candidate_reduced)

    child_motifs_reduced = []
    if node_data["expanded"]:
        children = sd.node_successors(node_id, compute=False)
        child_motifs_reduced = [
            sd.edge_stable_motif(node_id, s, reduced=True) for s in children
        ]

    child_motifs_bdd = state_list_to_bdd(symbolic_ctx, child_motifs_reduced)
    candidate_bdd = state_list_to_bdd(symbolic_ctx, candidate_states_reduced)

    children_set = ColoredVertexSet(symbolic_ctx, child_motifs_bdd)
    candidate_set = ColoredVertexSet(symbolic_ctx, candidate_bdd)

    avoid = candidate_set.union(children_set)

    if sd.config["debug"]:
        print(
            f"[{node_id}] > Start symbolic seed state identification with {len(candidate_states)} candidates and {avoid} avoid states."
        )

    seeds: list[BooleanSpace] = []
    sets: list[ColoredVertexSet] = []
    for i, candidate in enumerate(candidate_states_reduced):
        is_last = i == len(candidate_states_reduced) - 1
        is_minimal = len(child_motifs_reduced) == 0
        if seeds_only and is_minimal and is_last and len(seeds) == 0:
            # This node is pseudo-minimal, so it must contain at least
            # one attractor seed. If we only care about the attractor seeds,
            # we can thus return the last candidate without checking it.
            if sd.config["debug"]:
                print(
                    f"[{node_id}] > Single seed remaining in a (pseduo) minimal space. Done."
                )
            return ([candidate | node_space], None)
        candidate_singleton = graph_reduced.mk_subspace(candidate)

        avoid = avoid.minus(candidate_singleton)

        closure = symbolic_attractor_test(sd, node_id, graph_reduced, candidate, avoid)

        if closure is None:
            # This candidate can reach someone else in the candidate set,
            # or they can reach one of the child spaces. Hence it is not
            # an attractor.
            continue

        # Otherwise, we have an attractor set and we can save it.
        # It also becomes a part of the avoid set.
        avoid = avoid.union(closure)
        seeds.append(candidate | node_space)
        sets.append(closure)

    if sd.config["debug"]:
        print(f"[{node_id}] > Finished identification with {len(seeds)} seed states.")

    space_symbolic = sd.symbolic.mk_subspace(node_space).vertices()
    sets_converted: list[VertexSet] = []
    for s in sets:
        # Extend the attractor set with fixed nodes from the node space.
        vertices = s.vertices()
        vertices = sd.symbolic.transfer_from(vertices, graph_reduced)
        vertices = vertices.intersect(space_symbolic)
        sets_converted.append(vertices)

    return (seeds, sets_converted)


def symbolic_attractor_test(
    sd: SuccessionDiagram,
    node_id: int,
    graph: AsynchronousGraph,
    pivot: BooleanSpace,
    avoid_set: ColoredVertexSet,
) -> ColoredVertexSet | None:
    """
    Use symbolic reachability with saturation to compute the set of states
    reachable from the given `pivot`, or `None` if the reachability procedure
    intersects with any state in the `avoid` set.

    *The reason why we use `ColoredVertexSet` instead of `VertexSet` is mostly
    a technicality that should be irelevant in biobalm, since we don't allow any
    parameters outside of unknown inputs.*
    """

    if avoid_set.is_empty():
        avoid = None
    else:
        avoid = avoid_set

    reach_set = graph.mk_subspace(pivot)

    # The variables for which we already maintain that `reach_set`
    # contains all reachable states.
    saturated_vars: list[VariableId] = []

    # Variables where `reach_set` differs from the states in `avoid`.
    # We should prioritize updating these variables, because they *need*
    # to be updated if we are ever to reach `avoid`.
    conflict_vars: list[VariableId] = []

    # Populate conflict vars, assuming we have any.
    if avoid is not None:
        avoid_bdd = avoid.to_bdd()
        for name, val in pivot.items():
            var = graph.find_network_variable(name)
            bdd_var = graph.symbolic_context().find_network_bdd_variable(name)
            assert var is not None and bdd_var is not None
            incompatible = avoid_bdd.r_select({bdd_var: cast(Literal[0, 1], 1 - val)})
            if not incompatible.is_false():
                conflict_vars.append(var)
        conflict_vars = sort_variable_list(conflict_vars)
    all_conflict_vars = copy.copy(conflict_vars)

    # Remaining network variables that are still relevant, but may not
    # be necessary to reach `avoid`.
    other_vars: list[VariableId] = [
        x for x in graph.network_variables() if x not in conflict_vars
    ]
    other_vars = sort_variable_list(other_vars)

    if sd.config["debug"]:
        print(
            f"[{node_id}] > Start symbolic reachability with {len(conflict_vars)} conflict variables and {len(other_vars)} other variables."
        )

    # This algorithm has two modes of operation: If `avoid` is `None`,
    # then the only thing that we can do is compute forward reachability.
    # However, if we have any avoid states, we can run two reachability
    # operations that interleave each other: One forward from the
    # pivot vertex, the other backward from the `avoid` set. If these two
    # sets ever intersect, we know that there is a path from pivot into the
    # `avoid` set. To further speed up the computation, we first prioritize
    # the variables in which the pivot differs from the avoid set (so called
    # conflict variables). If all such variables are covered, we then consider
    # those that are close to the conflict variables in terms of regulations
    # (i.e. if we can't update a conflict variable, we ideally want to update
    # one that regulates it).

    # True if the main cycle completes with all saturation procedures fully
    # completed and no unprocessed variables remaining.
    all_done = False

    while not all_done:
        all_done = True

        # Saturate reach_set with currently selected variables, but only if
        # it's symbolic size is smaller than that of the avoid set (reach set
        # tends to grow quite large and we'd like to avoid that).
        saturation_done = False
        while not saturation_done:
            if avoid is not None and not avoid.intersect(reach_set).is_empty():
                if sd.config["debug"]:
                    print(f"[{node_id}] > Discovered avoid state. Done.")
                return None

            saturation_done = True
            for var in saturated_vars:
                successors = graph.var_post_out(var, reach_set)
                if not successors.is_empty():
                    all_done = False  # The main loop should continue.
                    updated = reach_set.union(successors)
                    no_avoid = avoid is None
                    avoid_is_larger = (
                        avoid is not None
                    ) and avoid.symbolic_size() >= updated.symbolic_size()
                    all_variables_done = (
                        len(conflict_vars) == 0 and len(other_vars) == 0
                    )
                    if no_avoid or avoid_is_larger or all_variables_done:
                        reach_set = updated
                        saturation_done = False
                        if reach_set.symbolic_size() > 100_000 and sd.config["debug"]:
                            print(
                                f"[{node_id}] > Saturation({len(saturated_vars)}) Incremented forward reach set: {reach_set}"
                            )
                        break

        if avoid is not None:
            # If `avoid` is not `None`, we also want to expand it backwards.
            saturation_done = False
            while not saturation_done:
                if not avoid.intersect(reach_set).is_empty():
                    if sd.config["debug"]:
                        print(f"[{node_id}] > Discovered avoid state. Done.")
                    return None

                saturation_done = True
                for var in saturated_vars:
                    predecessors = graph.var_pre_out(var, avoid)
                    if not predecessors.is_empty():
                        all_done = False
                        avoid = avoid.union(predecessors)
                        saturation_done = False
                        if avoid.symbolic_size() > 100_000 and sd.config["debug"]:
                            print(
                                f"[{node_id}] > Saturation({len(saturated_vars)}) Incremented backward avoid set: {avoid}"
                            )
                        break

        if sd.config["debug"] and (
            reach_set.symbolic_size() > 100_000
            or (avoid is not None and avoid.symbolic_size() > 100_000)
        ):
            print(
                f"[{node_id}] > Saturation({len(saturated_vars)}) Finished with avoid set {avoid} and reach set {reach_set}."
            )

        # Once saturated, try to expand the saturated
        # collection with either a conflict variable or
        # other variable.

        # First, we sort the non-conflict variables based on their
        # distance towards the conflict variables.
        network = sd.node_percolated_network(node_id)
        distances: dict[VariableId, int] = {
            v: network.variable_count() for v in network.variables()
        }
        visited: set[VariableId]
        if len(conflict_vars) > 0:
            visited = set(conflict_vars)
            current_level = set(conflict_vars)
        else:
            visited = set(all_conflict_vars)
            current_level = set(all_conflict_vars)

        next_level: set[VariableId] = set()

        distance = 0
        while len(current_level) > 0:
            for var in current_level:
                distances[var] = min(distance, distances[var])
                for s in network.predecessors(var):
                    if s not in visited:
                        visited.add(s)
                        next_level.add(s)
            current_level = next_level
            next_level = set()
            distance += 1

        other_sorted = sorted(other_vars, key=lambda x: distances[x])
        for var in conflict_vars + other_sorted:
            can_go_fwd = graph.var_post_out(var, reach_set)
            if avoid is not None:
                can_go_bwd = graph.var_pre_out(var, avoid)
            else:
                can_go_bwd = graph.mk_empty_colored_vertices()

            if can_go_fwd.is_empty() and can_go_bwd.is_empty():
                continue

            all_done = False

            reach_set = reach_set.union(can_go_fwd)
            if avoid is not None:
                avoid = avoid.union(can_go_bwd)

            if var in conflict_vars:
                conflict_vars.remove(var)
            if var in other_vars:
                other_vars.remove(var)

            saturated_vars.append(var)
            saturated_vars = sort_variable_list(saturated_vars)

            if sd.config["debug"]:
                print(
                    f"[{node_id}] > Saturation({len(saturated_vars)}) Added saturation variable. {len(conflict_vars)} conflict and {len(other_vars)} other variables remaining."
                )

            break

    if sd.config["debug"]:
        print(f"[{node_id}] > Reachability completed with {reach_set}.")

    return reach_set


def sort_variable_list(variables: list[VariableId]):
    return list(sorted(variables, reverse=True))



"""
Some basic utility operations on spaces (partial assignments of BN variables).

Each space is represented as a dictionary with a subset of variable names as
keys and values `0`/`1` assigned to fixed variables. See also
:class:`BooleanSpace<biobalm.types.BooleanSpace>`.
"""

from __future__ import annotations

from copy import copy
from typing import TYPE_CHECKING, Literal, cast

from biodivine_aeon import (
    AsynchronousGraph,
    BddVariableSet,
    BooleanNetwork,
    Percolation,
    SymbolicContext,
    UpdateFunction,
)

from biobalm.symbolic_utils import function_eval

if TYPE_CHECKING:
    from biodivine_aeon import BooleanExpression

    from biobalm.types import BooleanSpace


def intersect(x: BooleanSpace, y: BooleanSpace) -> BooleanSpace | None:
    """
    Compute the intersection of two spaces.

    Returns the intersection as a new space, or `None` if the spaces don't
    intersect.

    Parameters
    ----------
    x : BooleanSpace
        The first space.
    y : BooleanSpace
        The second space.

    Returns
    -------
    BooleanSpace | None
        The intersection of `x` and `y`, or `None` if the spaces don't
        intersect.
    """
    result: BooleanSpace = {}
    for k, v in x.items():
        result[k] = v
    for k, v in y.items():
        if k in result and result[k] != v:
            return None
        result[k] = v
    return result


def is_subspace(x: BooleanSpace, y: BooleanSpace) -> bool:
    """
    Checks if `x` is a subspace of `y`.

    Parameters
    ----------
    x : BooleanSpace
        The first space.
    y : BooleanSpace
        The second space.

    Returns
    -------
    bool
        `True` if `x` is a subspace of `y`.
    """
    for var in y:
        if var not in x:
            return False
        if x[var] != y[var]:
            return False
    return True


def dnf_function_is_true(dnf: list[BooleanSpace], state: BooleanSpace) -> bool:
    """
    Checks if a DNF function evaluates to `1` for the given state (or space).

    The DNF function is represented as a list of spaces, such that each space represents
    exactly one DNF clause.

    Parameters
    ----------
    dnf : list[BooleanSpace]
        The DNF function to evaluate.
    state : BooleanSpace
        The state in which the function is evaluated.

    Returns
    -------
    bool
        `True` if the function evaluates to `1` in the given state.
    """
    if len(dnf) == 0:
        return False

    for conjunction in dnf:
        if conjunction.items() <= state.items():
            return True
    return False


def remove_state_from_dnf(
    dnf: list[BooleanSpace], state: BooleanSpace
) -> list[BooleanSpace]:
    """
    Removes all clauses (conjunctions) that are `True` in the given `state` from a DNF function.

    The result is a new list (i.e., it does not modify the original list).

    Parameters
    ----------
    dnf : list[BooleanSpace]
        The DNF function to modify.
    state : BooleanSpace
        The state to remove from the function.

    Returns
    -------
    list[BooleanSpace]
        The modified DNF function.
    """
    modified_dnf: list[BooleanSpace] = []
    for conjunction in dnf:
        if conjunction.items() <= state.items():
            pass
        else:
            modified_dnf.append(conjunction)
    return modified_dnf


def percolate_space_strict(
    network: AsynchronousGraph, space: BooleanSpace
) -> BooleanSpace:
    """
    Percolates a space through a Boolean network, disregarding constants.

    Returns the set of variables which become fixed as a result of fixing the
    variables from `space` within the given `AsynchronousGraph`.

    Note that the strict percolation process does not propagate constants that
    are already fixed within the `network`, only those that are specified in
    `space`. Also, the result only contains any *new* constants, not those that
    are already fixed in `space`.

    Parameters
    ----------
    network : AsynchronousGraph
        A symbolic representation of the Boolean network (from which state
        transitions can be generated) in which the percolation is performed. If
        needed, an `AsynchronousGraph` object can be constructed from a
        `BooleanNetwork` via `biodivine_aeon.AsynchronousGraph(bn)`.
    space : BooleanSpace
        The space to percolate.

    Returns
    -------
    BooleanSpace
        The percolated space.
    """

    result: BooleanSpace = {}
    restriction: BooleanSpace = copy(space)
    candidates = set(network.network_variable_names())

    # Ignore variables that are already fixed.
    for var in network.network_variable_names():
        fn_bdd = network.mk_update_function(var)
        if fn_bdd.is_true() or fn_bdd.is_false():
            candidates.remove(var)

    done = False
    while not done:
        done = True
        for var in copy(candidates):
            fn_bdd = network.mk_update_function(var)
            fn_value = function_eval(fn_bdd, restriction)
            if fn_value is not None:
                if var in restriction and restriction[var] != fn_value:
                    # There is a conflict. We don't want to output this,
                    # but we also don't want to change the value.
                    candidates.remove(var)
                else:
                    done = False
                    restriction[var] = fn_value
                    result[var] = fn_value
                    candidates.remove(var)

    return result


def percolate_space(
    network: AsynchronousGraph,
    space: BooleanSpace,
) -> BooleanSpace:
    """
    Percolates a space through a Boolean network.

    Takes a symbolic `AsynchronousGraph` and a `BooleanSpace`. It then percolates
    any values that are effectively constant within the `network` assuming the variables
    from `space` are fixed accordingly.

    If the argument is a trap space, then the result is a subspace of the
    argument and is also a trap space.

    However, when the argument is a general space, the percolation can actually
    lead "outside" of the original space. In such case, the original fixed value
    is *not* modified and the conflict will remain in the resulting space.

    Parameters
    ----------
    network : AsynchronousGraph
        A symbolic representation of the Boolean network (from which state
        transitions can be generated) in which the percolation is performed. If
        needed, an `AsynchronousGraph` object can be constructed from a
        `BooleanNetwork` via `biodivine_aeon.AsynchronousGraph(bn)`.
    space : BooleanSpace
        The space to percolate.

    Returns
    -------
    BooleanSpace
        The percolated space.
    """

    percolated = Percolation.percolate_subspace(network, space)
    result: BooleanSpace = {}
    for var, value in percolated.items():
        var_name = network.get_network_variable_name(var)
        result[var_name] = cast(Literal[0, 1], int(value))
    return result


def percolation_conflicts(
    network: AsynchronousGraph,
    space: BooleanSpace,
    strict_percolation: bool = True,
) -> set[str]:
    """
    Find variables that conflict with the percolation of the given space.

    Returns a set of variables from `space` that are in conflict with the
    percolation of the given space (see `percolate_space`).

    Parameters
    ----------
    network : AsynchronousGraph
        A symbolic representation of the Boolean network (from which state
        transitions can be generated) in which the percolation is performed. If
        needed, an `AsynchronousGraph` object can be constructed from a
        `BooleanNetwork` via `biodivine_aeon.AsynchronousGraph(bn)`.
    space : BooleanSpace
        The space to percolate.
    strict_percolation : bool
        If `True` (the default), then the percolation is performed using
        :func:`percolate_space_strict`. Otherwise, it is performed using
        :func:`percolate_space`.

    Returns
    -------
    set[str]
        A set of variables from `space` that are in conflict with the
        percolation of the given space.
    """
    conflicts: set[str] = set()

    if strict_percolation:
        perc_space = percolate_space_strict(network, space)
    else:
        perc_space = percolate_space(network, space)

    for var, value in perc_space.items():
        fn_bdd = network.mk_update_function(var)
        fn_value = function_eval(fn_bdd, perc_space)
        if fn_value is not None and value != fn_value:
            conflicts.add(var)

    return conflicts


def percolate_network(
    bn: BooleanNetwork,
    space: BooleanSpace,
    symbolic_network: AsynchronousGraph | None = None,
    remove_constants: bool = False,
) -> BooleanNetwork:
    """
    Reduces a Boolean network by percolating a given space.

    Takes a `BooleanNetwork` and a `BooleanSpace`. It then produces a new network with
    update functions percolated based on the supplied space.

    The dynamics of the resulting network correspond to the dynamics of the
    network obtained by percolating the given space. If the space (or the
    percolated space) is a trap space, then the resulting dynamics are a
    subgraph of the original network's state transition graph. Otherwise, the
    dynamics correspond to the effects of an external intervention.

    The percolation process is based on BDD conversion. For this purpose, an optional
    `SymbolicContext` can be provided. If not given, a temporary `SymbolicContext` will
    be created instead. Note that this is necessary to resolve non-trivial tautologies or
    contradictions that can arise once the variables from `space` are fixed.

    Parameters
    ----------
    bn : BooleanNetwork
        The network to percolate.
    space : BooleanSpace
        The space to percolate.
    symbolic_network : AsynchronousGraph | None
        An optional symbolic representation to use to perform the percolation. If not
        given, a temporary one will be created from `bn`.
    remove_constants : bool
        If `True`, then the constants are removed from the resulting network. By
        default, `False`.

    Returns
    -------
    BooleanNetwork
        The percolated network.
    """

    if symbolic_network is None:
        symbolic_network = AsynchronousGraph(bn)
    var_set = symbolic_network.symbolic_context().bdd_variable_set()

    # Percolate the space first to ensure everything that can be fixed is fixed.
    space = percolate_space(symbolic_network, space)

    # Make a copy of the BN and copy the relevant functions.
    new_bn = copy(bn)

    for var in bn.variables():
        update = bn.get_update_function(var)
        if update is None:
            # This variable is a free input.
            assert len(bn.predecessors(var)) == 0
            name = bn.get_variable_name(var)
            if name in space:
                new_bn.set_update_function(
                    var, UpdateFunction.mk_const(new_bn, space[name])
                )
        else:
            percolated = restrict_expression(
                update.as_expression(), space, symbolic_context=var_set
            )
            new_update = UpdateFunction(new_bn, percolated)
            new_bn.set_update_function(var, new_update)

    new_bn = new_bn.infer_valid_graph()
    if remove_constants:
        new_bn = new_bn.inline_constants(infer_constants=True, repair_graph=True)

    return new_bn


def restrict_expression(
    expression: BooleanExpression,
    space: BooleanSpace,
    symbolic_context: BddVariableSet | SymbolicContext | None = None,
) -> BooleanExpression:
    """
    Restricts a Boolean expression to a given space.

    Takes a `BooleanExpression` and a `BooleanSpace`. Returns a simplified `BooleanExpression`
    that is valid for exactly the same members of the given `space` as the original expression.
    The resulting expression does not depend on the variables which are fixed in the given `space`.

    The percolation process is based on BDD conversion. For this purpose, an optional `BddVariableSet`
    can be provided. If not given, a temporary `BddVariableSet` will be created instead. Note that
    this is necessary to resolve non-trivial tautologies/contradictions that can arise once the
    variables from `space` are fixed.

    Parameters
    ----------
    expression : BooleanExpression
        The expression to restrict.
    space : BooleanSpace
        The space to restrict to.
    symbolic_context : BddVariableSet | SymbolicContext | None
        An optional symbolic context to use to perform the percolation. If not given,
        a temporary one will be created.

        This is a `biodivine_aeon.BddVariableSet` or a `biodivine_aeon.SymbolicContext`
        (which is automatically converted to `biodivine_aeon.BddVariableSet`).
        The context object ensures compatibility between BDDs by
        maintaining a shared collection of variable names and their ordering.

    Returns
    -------
    BooleanExpression
        The restricted expression.
    """

    variables = expression.support_set()
    space = {k: v for k, v in space.items() if k in variables}

    if len(space) == 0:
        return expression

    if symbolic_context is None:
        symbolic_context = BddVariableSet(sorted(variables))
    if isinstance(symbolic_context, SymbolicContext):
        symbolic_context = symbolic_context.bdd_variable_set()

    bdd = symbolic_context.eval_expression(expression)
    bdd = bdd.r_restrict(space)
    return bdd.to_expression()


def expression_to_space_list(
    expression: BooleanExpression,
    symbolic_context: BddVariableSet | SymbolicContext | None = None,
) -> list[BooleanSpace]:
    """
    Convert a Boolean expression to a list of subspaces for which it is true.

    Equivalent to a disjunctive normal form. Convert a `BooleanExpression` to a
    list of subspaces whose union represents an equivalent set of the network
    states which satisfy the expression.

    Note that the spaces are not necessarily pair-wise disjoint. Also, the list
    is not necessarily minimal.

    The translation uses a DNF conversion based on BDDs. For this purpose, an optional
    `BddVariableSet` can be provided. If not given, a temporary `BddVariableSet` will be
    created instead.

    Parameters
    ----------
    expression : BooleanExpression
        The expression to convert.
    symbolic_context : BddVariableSet | SymbolicContext | None
        An optional symbolic context to use to perform the percolation. If not given,
        a temporary one will be created.

        This is a `biodivine_aeon.BddVariableSet` or a `biodivine_aeon.SymbolicContext`
        (which is automatically converted to `biodivine_aeon.BddVariableSet`).
        The context object ensures compatibility between BDDs by
        maintaining a shared collection of variable names and their ordering.

    Returns
    -------
    list[BooleanSpace]
        The list of subspaces on which the expression is true.
    """

    if symbolic_context is None:
        variables = sorted(expression.support_set())
        symbolic_context = BddVariableSet(variables)
    if isinstance(symbolic_context, SymbolicContext):
        symbolic_context = symbolic_context.bdd_variable_set()

    bdd = symbolic_context.eval_expression(expression)

    sub_spaces: list[BooleanSpace] = []
    for clause in bdd.clause_iterator():
        space: BooleanSpace = {}
        for var, value in clause.items():
            space[symbolic_context.get_variable_name(var)] = cast(
                Literal[0, 1], int(value)
            )
        sub_spaces.append(space)

    return sub_spaces


def space_unique_key(space: BooleanSpace, network: BooleanNetwork) -> int:
    """
    Provide a unique hash key for the provided space in a given network.

    Computes an integer which is a unique representation of the provided `space`
    (with respect to the given `network`).

    This integer key can be used instead of the original `space` in places where
    dictionaries are not allowed, such as a key within a larger dictionary, or
    a sorting key.

    Note that when used for sorting, this key essentially implements a particular
    form of lexicographic ordering on spaces. This is always a total ordering
    (there is no ambiguity).

    Parameters
    ----------
    space : BooleanSpace
        The space to encode.
    network : BooleanNetwork
        The network in which the space is defined.

    Returns
    -------
    int
        A unique key for the space.
    """

    # Key is a binary encoding of the space dictionary. Since Python has
    # arbitrary-precision integers, this should work for any network and be
    # reasonably fast (we are not doing any copies or string manipulation).
    key: int = 0
    for k, v in space.items():
        var = network.find_variable(k)
        if var is None:
            raise IndexError(f"Unknown variable {var}.")
        # Each variable is encoded as two bits, so the total length
        # of the key is 2 * n and the offset of each variable is 2 * index.
        # 00 - unknown; 10 - zero; 11 - one
        key |= (v + 2) << (2 * int(var))
    return key



from __future__ import annotations

from typing import Literal, TypeAlias, TypedDict

import networkx as nx  # type: ignore
import biodivine_aeon as ba

BooleanSpace: TypeAlias = dict[str, Literal[0, 1]]
"""Type alias for `dict[str, Literal[0, 1]]`. Represents a Boolean subspace, which is defined by a set of fixed node values."""
SubspaceSuccession: TypeAlias = list[BooleanSpace]
"""Type alias for `list[BooleanSpace]`. Represents a sequence of subspaces, usually nested."""
ControlOverrides: TypeAlias = list[BooleanSpace]
"""Type alias for `list[BooleanSpace]`. Represents a sequence of subspaces, that percolate to targets."""


class SuccessionDiagramState(TypedDict):
    """
    A `TypedDict` class that stores the state of a succession diagram (see :class:`biobalm.SuccessionDiagram`).
    """

    network_rules: str
    """
    The network rules as an `.aeon` formatted string.
    """

    petri_net: nx.DiGraph
    """
    The Petri net representation of the network rules (see :mod:`biobalm.petri_net_translation`).
    """

    nfvs: list[str] | None
    """
    The negative feedback vertex set used for attractor detection.
    """

    dag: nx.DiGraph
    """
    The directed acyclic graph representation of the succession diagram structure, including
    the :class:`biobalm.types.NodeData` of each node.
    """

    node_indices: dict[int, int]
    """
    A dictionary mapping subspace keys to their positions in the succession
    diagram (see :func:`biobalm.space_utils.space_unique_key`).
    """

    config: SuccessionDiagramConfiguration
    """
    "Global" configuration of a succession diagram.
    """


class NodeData(TypedDict):
    """
    A `TypedDict` class that stores the data of a succession diagram node (see :class:`biobalm.SuccessionDiagram`).

    Returned from :func:`biobalm.SuccessionDiagram.node_data`.
    However, this class is not directly used at runtime, and only exists for static type-checking.
    Instead, at runtime, an untyped dictionary is used because that is what is returned by `networkx.DiGraph.nodes(data=True)`.
    """

    depth: int
    """
    The depth of the node in the succession diagram rooted directed acyclic
    graph. In cases of multiple paths from the root to the reference node,
    the longest is used.
    """

    space: BooleanSpace
    """
    The sub-space that the node represents (this subspace will always be a
    trap space).
    """

    expanded: bool
    """
    Whether the node has been expanded yet or not (`biobalm` builds the
    succession diagram lazily).

    If `expanded=False`, the node *must not* have any successor nodes.
    If `expanded=True`, the node must have all its successors computed
    and included in the succession diagram.
    """

    percolated_network: ba.BooleanNetwork | None
    """
    The AEON `BooleanNetwork` that has variables fixed and percolated
    according to the node `space`. Constant variables are then fully
    eliminated.

    If `None`, this has not been computed yet (`SuccessionDiagram.network` can be
    often used instead).
    """

    percolated_petri_net: nx.DiGraph | None
    """
    The Petri net representation of the network rules percolated to the
    node's sub-space (i.e. a Petri net encoding of the `percolated_network`).
    Constant variables are fully eliminated.

    If `None`, this has not been computed yet (`SuccessionDiagram.petri_net` can be
    often used instead).
    """

    percolated_nfvs: list[str] | None
    """
    An NFVS of the `percolated_network`.

    Note that this contains no variables that are fixed by the node `space`.
    Also, there is no guarantee that this is a subset of `SuccessionDiagram.nfvs`.

    If `None`, this has not been computed yet (`SuccessionDiagram.nfvs` can be
    often used instead).
    """

    attractor_candidates: list[BooleanSpace] | None
    """
    List of states that cover all network attractors in the node's sub-space
    (excluding child nodes, if known).

    Note that each complex/cyclic attractor can be covered by more than
    one state, but each attractor has to be covered by at least one state. Furthermore,
    outside of minimal trap spaces, some candidate states can cover zero attractors.
    That is, even if the candidate set is not empty, it does not guarantee that
    the non-minimal sub-space contains attractors.

    If `None`, these have not been computed, and the node may or may not
    have associated attractors.
    """

    attractor_seeds: list[BooleanSpace] | None
    """
    List of states that one-to-one correspond to network attractors in the node's
    sub-space (excluding child nodes, if known).

    This is very similar to `attractor_candidates`, but here, it is guaranteed
    that each attractor is represented by exactly one state.

    If `None`, these have not been computed, and the node may or may not
    have associated attractors.
    """

    attractor_sets: list[ba.VertexSet] | None
    """
    List of attractors that are present in the node's sub-space (excluding
    child nodes, if known).

    Each attractor is represented symbolically using `biodivine_aeon.VertexSet`.

    Note that for fixed-point attractors, the attractor set is effectively
    equivalent to the attractor seed. However, for complex attractors, this
    set containes *all* the attractor states as opposed to just one. Hence it
    is harder to compute, but can be used to analyze things like the average
    value of each variable in the attractor states, or the presence of
    particular oscillation patterns.

    If `None`, these have not been computed, and the node may or may not
    have associated attractors.
    """

    parent_node: int | None
    """
    The ID of one of the predecessor nodes through which this node was first
    discovered, or `None` if the node was discovered through other means
    (usually either the root, or a minimal trap space that was added without
    expanding its predecessors).
    """

    skipped: bool | None
    """
    If `True`, indicates that the successors of this node are not the normal
    maximal trap spaces, but rather some smaller trap spaces deeper in the
    succession diagram. The only requirement is that all minimal trap spaces
    that are reachable from this "skip node" are still reachable through
    this modified set of successors (in the most basic form, the successors
    can be the minimal trap spaces themselves).

    For skip nodes, attractor detection still works, but can over-count
    motif-avoidant attractors assuming the attractor intersects multiple
    skip nodes. In other words, each attractor is still found, but it is
    not necessarily true that its smallest trap space is discovered.
    Nevertheless, if the network has no motif-avoidant attractors, the
    result is always the same regardless of how many skip nodes are used.

    A skip node is considered "expanded", since its outgoing edges are
    computed.
    """


class SuccessionDiagramConfiguration(TypedDict):
    """
    Describes the configuration options of a `SuccessionDiagram`.

    Use :meth:`SuccessionDiagram.default_config` to create a
    configuration dictionary pre-populated with default values.
    """

    debug: bool
    """
    If `True`, the `SuccessionDiagram` will print messages
    describing the progress of the running operations.

    [Default: False]
    """

    max_motifs_per_node: int
    """
    Limit on the number of stable motifs explored for one node of a succession
    diagram. If this limit is exceeded during node expansion, a `RuntimeError`
    is raised and the node remains unexpanded.

    This limit is in place mainly to avoid surprising out of memory errors,
    because currently there is no logging mechanism that would report the
    number of stable motifs gradually.

    [Default: 100_000]
    """

    nfvs_size_threshold: int
    """
    For networks larger than this threshold, we only run FVS detection
    instead of NFVS detection. This is still correct, but can produce
    a larger node set.


    There is a trade-off between the speed gains from a smaller node set
    to consider and the cost of determining which FVS nodes only
    intersect negative cycles to find an NFVS subset. Typically,
    for smaller networks, the trade-off is in favor of
    computing a smaller NFVS.
    """

    pint_goal_size_limit: int
    """
    Pint is called using command line and can only accept a limited number of arguments.
    This limit is applied when constructing goals to avoid reaching this argument limit.

    The limit currently applies to the total number of literals that can be used to
    represent a goal.

    The default value was empirically tested as safe on Debian linux, but other operating
    systems may need a different limit to stay safe. Nevertheless, this should not be
    an issue on smaller/simpler networks.
    """

    attractor_candidates_limit: int
    """
    If more than `attractor_candidates_limit` states are produced during the
    attractor detection process, then the process fails with a `RuntimeError`.
    This is mainly to avoid out-of-memory errors or crashing `clingo`.
    """

    retained_set_optimization_threshold: int
    """
    If there are more than this amount of attractor candidates, the attractor
    detection process will try to optimize the retained set using ASP (if enabled).
    """

    minimum_simulation_budget: int
    """
    The minimum number of simulation steps per network variable that is guaranteed to be
    spent on eliminating attractor candidate states. That is, if budget is `1_000` and
    the network has `200` variables, the total number of allowed
    simulation steps is `200_000`.

    Note that this is a budget that applies to all candidates collectively. So if the number
    of candidates is larger, the number of steps per candidate is proportionally smaller.
    However, this budget only applies when simulation has not been able to make progress
    in the recent round. That is, if simulation has actively eliminated some candidates in
    the recent round, it will still continue regardless of the budget limit.
    """
